<?php
/**
 * Provides compatibility with All in One SEO
 * 
 * @package  Page Generator Pro
 * @author   Tim Carr
 * @version  2.9.0
 */
class Page_Generator_Pro_AIOSEO extends Page_Generator_Pro_Integration {

    /**
     * Holds the base object.
     *
     * @since   2.9.0
     *
     * @var     object
     */
    public $base;

    /**
     * Constructor
     * 
     * @since   2.9.0
     *
     * @param   object $base    Base Plugin Class
     */
    public function __construct( $base ) {

        // Store base class
        $this->base = $base;

        // Set Plugin
        $this->plugin_folder_filename = array(
            'all-in-one-seo-pack/all_in_one_seo_pack.php',
            'all-in-one-seo-pack-pro/all_in_one_seo_pack.php',
        );

        // Add Overwrite Section if All in One SEO enabled
        add_filter( 'page_generator_pro_common_get_content_overwrite_sections', array( $this, 'add_overwrite_section' ) );

        // Ignore All in One SEO meta keys if overwriting is disabled for All in One SEO
        add_filter( 'page_generator_pro_generate_set_post_meta_ignored_keys', array( $this, 'prevent_post_meta_copy_to_generated_content' ), 10, 4 );

    }

    /**
     * Defines available content overwrite sections.
     *
     * @since   2.9.0
     *
     * @param   array   $sections    Content Overwrite Sections
     * @return  array                Content Overwrite Sections
     */
    public function add_overwrite_section( $sections ) {

        // Bail if All in One SEO isn't active
        if ( ! $this->is_active() ) {
            return $sections;
        }

        // Add All in One SEO
        $sections['aioseo'] = __( 'All in One SEO', 'page-generator-pro' );

        // Return
        return $sections;

    }

    /**
     * Adds All in One SEO Post Meta Keys to the array of excluded Post Meta Keys if All in One SEO
     * metadata should not be overwritten on Content Generation
     *
     * @since   2.9.0
     *
     * @param   array   $ignored_keys   Ignored Keys (preg_match() compatible regex expressions are supported)
     * @param   int     $post_id        Generated Post ID
     * @param   array   $settings       Group Settings
     * @param   array   $post_args      wp_insert_post() / wp_update_post() arguments
     * @return  array   $ignored_keys   Ignored Keys
     */
    public function prevent_post_meta_copy_to_generated_content( $ignored_keys, $post_id, $settings, $post_args ) {

        // Determine if we want to create/replace All in One SEO metdata
        $overwrite = ( isset( $post_args['ID'] ) && ! array_key_exists( 'aioseo', $settings['overwrite_sections'] ) ? false : true ); 

        // If overwriting is enabled, no need to exclude anything
        if ( $overwrite ) {
            // Overwrite in AIOSEO DB for each Generated Page
            add_action( 'page_generator_pro_generate_content_finished', array( $this, 'update_aioseo_post_table' ), 10, 5 );

            // Return
            return $ignored_keys;
        }

        // Add All in One SEO Meta Keys so they are not overwritten on the Generated Post
        return array_merge( $ignored_keys, array(
            "/^_aioseop_(.*)/i"
        ) );

    }

    /**
     * Update AIOSEO DB table values for the Generated Post
     *
     * @since   2.9.5
     *
     * @param   int     $post_id        Post ID
     * @param   int     $group_id       Group ID
     * @param   array   $settings       Group Settings
     * @param   int     $index          Keyword Index
     * @param   bool    $test_mode      Test Mode
     */
    public function update_aioseo_post_table( $post_id, $group_id, $settings, $index, $test_mode ) {

        // Bail if AIOSEO Pro v4 doesn't exist
        if ( ! class_exists( 'AIOSEO\Plugin\Common\Models\Post' ) ) {
            return;
        }

        // Bail if AIOSEO Post Meta doesn't exist
        if ( ! isset( $settings['post_meta']['_aioseo_title'] ) ) {
            return;
        }

        // Define AIOSEO Post Table columns that we can fill from the generated Post's meta
        $columns = array(
            'title',
            'description',
            'keywords',
            'og_title',
            'og_description',
            'og_article_section',
            'og_article_tags',
            'twitter_title',
            'twitter_description',
        );

        // Build AIOSEO Post Table data
        $data = array(
            'id'        => $post_id,
            'context'   => 'post',
        );
        foreach ( $columns as $column ) {
            $data[ $column ] = ( ( isset( $settings['post_meta']['_aioseo_' . $column ] ) && ! empty( $settings['post_meta']['_aioseo_' . $column ] ) ) ? $settings['post_meta']['_aioseo_' . $column ] : null );
        }

        // Update
        AIOSEO\Plugin\Common\Models\Post::savePost( $post_id, $data );

    }

 }