<?php
/**
 * Divi Module
 *
 * @package Page_Generator_Pro
 * @author  Tim Carr
 * @version 3.0.7
 */
class Page_Generator_Pro_Divi_Module extends ET_Builder_Module {

	/**
	 * How modules are supported in the Visual Builder (off|partial|on)
	 *
	 * @since 	3.0.7
	 *
	 * @var 	string
	 */
	public $vb_support = 'partial';

	/**
	 * Holds the shortcode properties
	 *
	 * @since 	3.0.7
	 *
	 * @var 	array
	 */
	private $shortcode;

	/**
	 * Defines the Module name
	 *
	 * @since 	3.0.7
	 */
	public function init() {

		// Get shortcode
		$this->shortcode = Page_Generator_Pro()->get_class( 'shortcode' )->get_shortcode( str_replace( 'page-generator-pro-divi-', '', $this->slug ) );
		if ( ! $this->shortcode ) {
			return;
		}

		// Define name
		$this->name = '_' . $this->shortcode['title'];

	}
	
	/**
	 * Defines the fields that can be configured for this Module
	 *
	 * @since 	3.0.7
	 */
	public function get_fields() {

		// Bail if no shortcode
		if ( ! $this->shortcode ) {
			return array();
		}

		// Build fields
		$fields = array();
		foreach ( $this->shortcode['fields'] as $field_name => $field ) {
			// Start building field definition
			$fields[ $field_name ] = array(
				'type'			=> $field['type'],
				'default'		=> ( isset( $field['default_value'] ) ? $field['default_value'] : '' ),
				'description'   => ( isset( $field['description'] ) ? $field['description'] : '' ),
				'label'         => $field['label'],
				'toggle_slug'   => $this->get_tab( $field_name ),
			);

			// Add field condition, if defined
			if ( isset( $field['condition'] ) ) {
				// Define comparison (show_if or show_if_not)
				$comparison = ( ( $field['condition']['comparison'] == '==' ) ? 'show_if' : 'show_if_not' );
				
				// Define value as 'on' or 'off' if it's 1 or 0 and the comparison field is a toggle field
				$value = $field['condition']['value'];
				if ( $this->shortcode['fields'][ $field['condition']['key'] ]['type'] == 'toggle' ) {
					$value = ( ($value == '1' ) ? 'on' : 'off' );
				}

				// Add the comparison condition
				$fields[ $field_name ][ $comparison ] = array(
					$field['condition']['key'] => $value,
				);
			}

			// Add/change field parameters depending on the field's type
			switch ( $field['type'] ) {
				/**
				 * Autocomplete
				 */
				case 'autocomplete':
				case 'text_multiple':
				case 'select_multiple':
					$fields[ $field_name ]['type'] = 'text'; // @TODO add support for these field types in Divi
					break;

				/**
				 * Number
				 */
				case 'number':
					$fields[ $field_name ] = array_merge( $fields[ $field_name ], array(
						'type' 		=> 'range',
						'range_settings' 	=> array(
							'min' 	=> $field['min'],
							'max' 	=> $field['max'],
							'step' 	=> $field['step'],
						),
						'unitless'        => true,
					) );
					break;

				/**
				 * Select
				 */
				case 'select':
					$fields[ $field_name ]['options'] = $field['values'];
					break;

				/**
			     * Toggle
			     */
				case 'toggle':
					$fields[ $field_name ] = array_merge( $fields[ $field_name ], array(
						'type' 		=> 'yes_no_button',
						'default'	=> ( $fields[ $field_name ]['default'] ? 'on' : 'off' ),
						'options' 	=> array(
							'off' 		=> __( 'No', 'page-generator-pro' ),
							'on' 		=> __( 'Yes', 'page-generator-pro' ),
						),
					) );
					break;


			}
		}

		// Return
		return $fields;

	}

	/**
	 * Define tabs for the Module
	 *
	 * @since 	3.0.7
	 */
	public function get_settings_modal_toggles() {

		// Build tabs as toggles
		$toggles = array();
		$priority = 24;
		foreach ( $this->shortcode['tabs'] as $tab_name => $tab_properties ) {
			$toggles[ $tab_name ] = array(
				'priority' 	=> $priority,
				'title' 	=> $tab_properties['label'],
			);

			$priority++;
		}

		// Return
		$modal_toggles = array(
			'advanced' => array(
				'toggles' => $toggles,
			),
		);
		
		return $modal_toggles;

	}

	/**
	 * Renders the shortcode syntax, converted from the module's properties array
	 *
	 * @since 	3.0.7
	 */
	public function render( $unprocessed_props, $content = null, $render_slug ) {

		// Bail if no shortcode
		if ( ! $this->shortcode ) {
			return '';
		}

		// Return [shortcode] string
		return Page_Generator_Pro()->get_class( 'shortcode' )->build_shortcode(
			Page_Generator_Pro()->plugin->name . '-' . str_replace( 'page-generator-pro-divi-', '', $this->slug ),
			$this->shortcode['fields'],
			$this->props
		);
		
	}

	/**
	 * Returns the Tab Name for the given Field Name
	 *
	 * @since 	3.0.7
	 *
	 * @param 	string 	$field_name 	Field Name
	 * @return 	string 					Tab Name
	 */
	private function get_tab( $field_name ) {

		foreach ( $this->shortcode['tabs'] as $tab_name => $tab_properties ) {
			if ( in_array( $field_name, $tab_properties['fields'] ) ) {
				return $tab_name;
			}
		}

		// If here, no tab could be found for this field to be rendered within,
		// so return Divi's main_content tab
		return 'main_content';

	}

}