<?php
/**
 * Genesis class
 * 
 * @package Page Generator Pro
 * @author  Tim Carr
 * @version 3.0.5
 */
class Page_Generator_Pro_Genesis extends Page_Generator_Pro_Integration {

    /**
     * Holds the base object.
     *
     * @since   3.0.5
     *
     * @var     object
     */
    public $base;

    /**
     * Constructor
     *
     * @since   3.0.5
     *
     * @param   object $base    Base Plugin Class
     */
    public function __construct( $base ) {

        // Store base class
        $this->base = $base;

        // Register Post Type Support for Content Groups
        add_action( 'init', array( $this, 'register_post_type_support' ) );

        // Add Overwrite Section if All in One SEO enabled
        add_filter( 'page_generator_pro_common_get_content_overwrite_sections', array( $this, 'add_overwrite_section' ) );

        // Ignore All in One SEO meta keys if overwriting is disabled for All in One SEO
        add_filter( 'page_generator_pro_generate_set_post_meta_ignored_keys', array( $this, 'prevent_post_meta_copy_to_generated_content' ), 10, 4 );

    }

    /**
     * Register Genesis Meta Boxes as supported for the Content Groups Post Type
     *
     * @since   3.0.5
     */
    public function register_post_type_support() {

        // Bail if Genesis isn't active
        if ( ! $this->is_active() ) {
            return;
        }

        // Add Support
        add_post_type_support( 'page-generator-pro', array(
            'genesis-seo',
            'genesis-scripts',
            'genesis-layouts',
            'genesis-breadcrumbs-toggle',
            'genesis-footer-widgets-toggle',
            'genesis-title-toggle'
        ) );

    }

    /**
     * Defines available content overwrite sections.
     *
     * @since   3.0.5
     *
     * @param   array   $sections    Content Overwrite Sections
     * @return  array                Content Overwrite Sections
     */
    public function add_overwrite_section( $sections ) {

        // Bail if Genesis isn't active
        if ( ! $this->is_active() ) {
            return $sections;
        }

        // Add Genesis
        $sections['genesis'] = __( 'Genesis Framework', 'page-generator-pro' );

        // Return
        return $sections;

    }

    /**
     * Adds Genesis Post Meta Keys to the array of excluded Post Meta Keys if Genesis
     * metadata should not be overwritten on Content Generation
     *
     * @since   3.0.5
     *
     * @param   array   $ignored_keys   Ignored Keys (preg_match() compatible regex expressions are supported)
     * @param   int     $post_id        Generated Post ID
     * @param   array   $settings       Group Settings
     * @param   array   $post_args      wp_insert_post() / wp_update_post() arguments
     * @return  array   $ignored_keys   Ignored Keys
     */
    public function prevent_post_meta_copy_to_generated_content( $ignored_keys, $post_id, $settings, $post_args ) {

        // Determine if we want to create/replace All in One SEO metdata
        $overwrite = ( isset( $post_args['ID'] ) && ! array_key_exists( 'genesis', $settings['overwrite_sections'] ) ? false : true ); 

        // If overwriting is enabled, no need to exclude anything
        if ( $overwrite ) {
            return $ignored_keys;
        }

        // Add Genesis Meta Keys so they are not overwritten on the Generated Post
        return array_merge( $ignored_keys, array(
            "/^_genesis_(.*)/i"
        ) );

    }

    /**
     * Checks if Genesis is active
     *
     * @since   3.0.5
     *
     * @return  bool    Is Active
     */
    public function is_active() {

        return defined( 'GENESIS_LOADED_FRAMEWORK' );
        
    }

}