<?php
/**
 * Integration class.  Used by integration classes for common functions,
 * such as checking if a Plugin is active, fetching the Plugin version
 * and whether a Plugin version meets the minimum required.
 * 
 * @package Page_Generator_Pro
 * @author  Tim Carr
 * @version 3.0.5
 */
class Page_Generator_Pro_Integration {

    /**
     * Holds the Plugin Folder and Filename
     *
     * @since   3.0.5
     *
     * @var     mixed   bool | string
     */
    public $plugin_folder_filename = false;

    /**
     * Holds the minimum supported version
     *
     * @since   3.0.5
     *
     * @var     mixed   bool | decimal
     */
    public $minimum_supported_version = false;

    /**
     * Checks if the Plugin is active, and if it meets the minimum supported
     * version, if specified.
     *
     * @since   3.0.5
     *
     * @return  bool    Plugin is Active
     */
    public function is_active() {

        // Bail if no Plugin Filename setup
        if ( ! is_array( $this->plugin_folder_filename ) && ! $this->plugin_folder_filename ) {
            return false;
        }

        // Assume Plugin isn't active if we can't use WordPress' function
        if ( ! function_exists( 'is_plugin_active' ) ) {
            return false;
        }

        // Assume Plugin isn't active
        $is_active = false;

        // If our Plugin Filename is array, iterate through it
        if ( is_array( $this->plugin_folder_filename ) ) {
            foreach ( $this->plugin_folder_filename as $plugin_folder_filename ) {
                if ( is_plugin_active( $plugin_folder_filename ) ) {
                    $is_active = true;
                    break;
                }
            }
        } else {
            $is_active = is_plugin_active( $this->plugin_folder_filename );
        }

        // Bail if Plugin isn't active
        if ( ! $is_active ) {
            return false;
        }

        // If there's no minimum supported version, the Plugin is active
        if ( ! $this->minimum_supported_version ) {
            return true;
        }

        // If the Plugin doesn't match the minimum supported version, deem it as not active
        if ( $this->get_version() < $this->minimum_supported_version ) {
            return false;
        }

        // If here, the Plugin is active and meets the minimum supported version
        return true;

    }

    /**
     * Returns the Integration Plugin's version
     *
     * @since   3.0.5
     *
     * @return  decimal     Version
     */
    public function get_version() {

        $plugin_data = get_file_data( WP_PLUGIN_DIR . '/' . $this->plugin_folder_filename, array( 'Version' => 'Version' ), false );
        return $plugin_data['Version'];
        
    }

}