<?php
/**
 * Provides compatibility with WPSSO
 * 
 * @package  Page Generator Pro
 * @author   Tim Carr
 * @version  3.2.0
 */
class Page_Generator_Pro_WPSSO extends Page_Generator_Pro_Integration {

    /**
     * Holds the base object.
     *
     * @since   3.2.0
     *
     * @var     object
     */
    public $base;

    /**
     * Constructor
     * 
     * @since   3.2.0
     *
     * @param   object $base    Base Plugin Class
     */
    public function __construct( $base ) {

        // Store base class
        $this->base = $base;

        // Set Plugin
        $this->plugin_folder_filename = 'wpsso/wpsso.php';

        // Register support fpr WPSSO
        add_filter( 'sucom_get_post_types', array( $this, 'register_support' ) );

        // Add Overwrite Section if WPSSO enabled
        add_filter( 'page_generator_pro_common_get_content_overwrite_sections', array( $this, 'add_overwrite_section' ) );

        // Ignore WPSSO meta keys if overwriting is disabled for WPSSO
        add_filter( 'page_generator_pro_generate_set_post_meta_ignored_keys', array( $this, 'prevent_post_meta_copy_to_generated_content' ), 10, 4 );
    
    }

    /**
     * Registers Content Groups with WPSSO, so that WPSSO's metaboxes are output
     * on Content Groups
     *
     * @since   3.2.0
     *
     * @param   array   $post_types     Post Types
     * @return  array                   Post Types
     */
    public function register_support( $post_types ) {

        $post_types['page-generator-pro'] = 'page-generator-pro';
        return $post_types;

    }

    /**
     * Defines available content overwrite sections.
     *
     * @since   3.2.0
     *
     * @param   array   $sections    Content Overwrite Sections
     * @return  array                Content Overwrite Sections
     */
    public function add_overwrite_section( $sections ) {

        // Bail if WPSSO isn't active
        if ( ! $this->is_active() ) {
            return $sections;
        }

        // Add WPSSO
        $sections['wpsso'] = __( 'WPSSO', 'page-generator-pro' );

        // Return
        return $sections;

    }

    /**
     * Adds WPSSO Post Meta Keys to the array of excluded Post Meta Keys if WPSSO
     * metadata should not be overwritten on Content Generation
     *
     * @since   3.2.0
     *
     * @param   array   $ignored_keys   Ignored Keys (preg_match() compatible regex expressions are supported)
     * @param   int     $post_id        Generated Post ID
     * @param   array   $settings       Group Settings
     * @param   array   $post_args      wp_insert_post() / wp_update_post() arguments
     * @return  array   $ignored_keys   Ignored Keys
     */
    public function prevent_post_meta_copy_to_generated_content( $ignored_keys, $post_id, $settings, $post_args ) {

        // Determine if we want to create/replace WPSSO metdata
        $overwrite = ( isset( $post_args['ID'] ) && ! array_key_exists( 'wpsso', $settings['overwrite_sections'] ) ? false : true ); 

        // If overwriting is enabled, no need to exclude anything
        if ( $overwrite ) {
            return $ignored_keys;
        }

        // Add WPSSO Meta Keys so they are not overwritten on the Generated Post
        return array_merge( $ignored_keys, array(
            "/^_wpsso_(.*)/i"
        ) );

    }

 }